/******************************************************************************
 * cached_objects_hash.c
 *  
 * Author:
 * Stelios Sidiroglou-Douskos <stelios@cs.columbia.edu>
 * Department of Computer Science - Columbia University
 *
 * Copyright (c) 2003 by Columbia University, all rights reserved.
 * ***************************************************************************/


/*Glib libs*/
#include <pthread.h>            /* Threading lib */
#define _GNU_SOURCE
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <glib.h>
#include <string.h>

#include "global.h"
#include "md5.h"
#include "main.h"
#include "cached_objects_hash.h"


GHashTable *cached_objects_hash;
GHashTable *reverse_cache_hash;
FILE *cob_log_file;

/*
 * cached_objects_hash_init()
 * Initialize hash table
 */
void cached_objects_hash_init(void)
{
	 cached_objects_hash = NULL;
	 reverse_cache_hash = NULL;
	 
	 cached_objects_hash = g_hash_table_new(g_str_hash, g_str_equal);
	 reverse_cache_hash = g_hash_table_new(g_str_hash, g_str_equal);
}

/*
 * cache_objects_hash_destroy()
 * Destroy hash table, free memory
 */
void cache_objects_hash_destroy(void)
{
	g_hash_table_destroy(cached_objects_hash);
	g_hash_table_destroy(reverse_cache_hash);
}


/*
 * cob_insert_report()
 * Insert the report into hash table
 *
void cob_insert_report(report_t *report)
{
	if (report == NULL) {
		fprintf(stderr, "cob_insert_report:: Report was empty!\n");
		return;
	}
	cached_objects_hash =  g_slist_append(cached_objects_hash, report);
}
*/

/*
 * is_url_in_hash()
 * Checks to see if url is in the hash table
 * Returns pointer if it is, NULL otherwise
 */
gpointer is_url_in_hash(char *url)
{
	gpointer value = NULL;
    pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
	
	pthread_mutex_lock(&mutex);
    value = g_hash_table_lookup(cached_objects_hash,      /*table */
                                url);      		  /*key */
    pthread_mutex_unlock(&mutex);


    if (value == NULL)
        return NULL;
    else
        printf("HASH LOOKUP IS %s\n", (char *)value);

	return value;
}

/*
 * is_path_in_hash()
 * If the path has an associated url, that url is returned.
 * 
 * Returns pointer if it is, NULL otherwise
 */
gpointer is_path_in_hash(char *path)
{
	gpointer value = NULL;
    pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
	
	pthread_mutex_lock(&mutex);
    value = g_hash_table_lookup(reverse_cache_hash,   /*table */
                               	path);      		  /*key */
    pthread_mutex_unlock(&mutex);


    if (value == NULL)
        return NULL;
    else
        printf("PATH LOOKUP IS %s\n", (char *)value);

	return value;

}

/*
 * cob_put_in_cache()
 * Places the path of the file under the url key
 */
void cob_put_in_cache(char *path, char *url)
{
	pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;

    printf("Inserting -> url %s\n path%s\n", url, path);

    pthread_mutex_lock(&mutex);
    g_hash_table_insert(cached_objects_hash, url, path);
	g_hash_table_insert(reverse_cache_hash, path, url);
    pthread_mutex_unlock(&mutex);
}

/*
 * cob_get_number_in_cache()
 * Returns the number of elements in the hash
 */
unsigned int cob_get_number_in_cache(void)
{
	pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
	unsigned int size;
	 
	pthread_mutex_lock(&mutex);
    size = g_hash_table_size(cached_objects_hash);
	pthread_mutex_unlock(&mutex);
	
	return size;
}

/*
 *  cob_remover_from_cache()
 *  Remove key from hash
 */
void cob_remove_from_cache(char *referer)
{
	pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
	pthread_mutex_lock(&mutex);
	g_hash_table_remove(cached_objects_hash, referer);	
	pthread_mutex_unlock(&mutex);
}

/*
 * cob_foreach_func()
 * 
 * Goes through every key value pair and logs them to a file
 */
void cob_foreach_func(gpointer key, gpointer value, gpointer user_data)
{
	//FILE *cob_log_file;
	
	//cob_log_file = fopen("LOG_COB", "w");
	
	if (cob_log_file == NULL) {
		fprintf(stderr, "cob_foreach_func:: couldn't open log file \n");
		return;
	}
	
	fprintf(cob_log_file, "%s$%s\n", (char *)key, (char *)value);
	printf("foreach:%s:%s\n", (char *)key, (char *)value);
}

/*
 * cob_log_data()
 * 
 * Function to initiate the logging of the cache buffer
 */
void cob_log_data()
{
	pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
	pthread_mutex_lock(&mutex);

	printf("cob_log_data::\n");
	cob_log_file = fopen("LOG_COB", "w");
	if (cob_log_file != NULL)
			g_hash_table_foreach(cached_objects_hash, cob_foreach_func, NULL);

	fflush(cob_log_file);
	fclose(cob_log_file);
	pthread_mutex_unlock(&mutex);

}

/*
 * cob_create_from_log()
 * Create the hash table from the log file
 */
void cob_create_from_log()
{	
	size_t read;
	char *line = NULL;
	size_t len = 0;
	char *key;
	char *value;
	
	pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
	
	pthread_mutex_lock(&mutex);
	printf("cob_create_from_log::\n");

	cob_log_file = fopen("LOG_COB", "r");

	if (cob_log_file == NULL) {
		fprintf(stderr, "cob_create_from_log:: couldn't open log file \n");
		return;
	}

	while ((read = getline(&line, &len, cob_log_file)) != -1) {
		key = strtok(line, "$");
		printf("t:%s\n", key);
		value = strtok(NULL, "$");
		printf("t2:%s\n", value);

		if ((key != NULL) && (value != NULL)) {
			cob_put_in_cache(value, key);
		} else {
			fprintf(stderr, "cob_create_from_log:: key or value was NULL"
							" skipping insertion\n");
		}
	}	
	
	fclose(cob_log_file);
	pthread_mutex_unlock(&mutex);
}
