/******************************************************************************
 * email_objects_buffer.c
 *  
 * Author:
 * Stelios Sidiroglou-Douskos <stelios@cs.columbia.edu>
 * Department of Computer Science - Columbia University
 *
 * Copyright (c) 2003 by Columbia University, all rights reserved.
 * ***************************************************************************/

/*Glib libs*/
#include <pthread.h>            /* Threading lib */
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <glib.h>
#include <string.h>
#include <libesmtp.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>

#include "main.h"

/* Upload mechanism */
enum {FIFO, LIFO, SJF};

int upload_mechanism = SJF;

/*
 * Doubly linked list holding the email objects
 */
GList *email_objects_buffer;

/*
 * email_objects_buffer_init()
 *
 * Initialize the list
 */
void email_objects_buffer_init(void)
{
	email_objects_buffer = NULL;
}

/*
 * email_get_objects()
 *
 * Returns the linked list
 */
GList *email_get_objects(void)
{
	return email_objects_buffer;
}

/*
 * email_append_object()
 *
 * Inserts object to linked list
 */
void email_append_object(email_object_t *object)
{
	printf("email_append_object: Adding email from %s "
					" to %s with md5 %s ", 
					object->mail_from, 
					object->mail_to,
					object->md5);
	email_objects_buffer = g_list_append(email_objects_buffer, object);
}

/*
 * delete_this_email()
 */
static void delete_this_email(email_object_t *object)
{
	email_objects_buffer = g_list_remove(email_objects_buffer, object);
}

/*
 * email_delete_object()
 */
void email_delete_object(char *md5)
{
	GList *iterator;
	email_object_t *email_tmp;
	
	for (iterator = email_objects_buffer; iterator != NULL; 
								iterator = iterator->next) {
		email_tmp = (email_object_t *) iterator->data;
		printf("MD5 is : %s\n", email_tmp->md5);
		if (strcmp(email_tmp->md5, md5) == 0) {
			delete_this_email(email_tmp);
		}
	}
}

/*
 * email_compare_size_func()
 *
 * Compares the size of the objects and reports the in Ascending size
 * order.
 */
gint email_compare_size_func(email_object_t *a , email_object_t *b)
{
	return a->data_size - b->data_size;
}


/*
 * email_object_init()
 */
static void email_object_init(email_object_t *object)
{
	bzero(object->mail_from, sizeof(object->mail_from));
	bzero(object->mail_to, sizeof(object->mail_to));
	bzero(object->body, sizeof(object->body));
	bzero(object->md5, sizeof(object->md5));
	object->rcpt_to = NULL;
	object->data = NULL;
	object->data_size = 0;
}

/*
 * email_create_object()
 *
 *  Returns object if succesful, NULL otherwise
 */
email_object_t *email_create_object()
{
	email_object_t *object = (email_object_t *) malloc(sizeof(email_object_t));
	
	if (!object) {
		fprintf(stderr, "email_create_object: No mem for new object\n");
	} else {
		email_object_init(object);
	}

	return object;
}

/*
 * email_show_objects()
 *
 * Walks through email objects and prints information
 */
void email_show_objects()
{
	GList *iterator;
	email_object_t *email_tmp;
	char *mail_to;

	printf("\n Dumping email objects list...\n");
	for (iterator = email_objects_buffer; iterator != NULL; 
								iterator = iterator->next) {
		email_tmp = (email_object_t *) iterator->data;
		mail_to = email_tmp->mail_to;
		printf("Mail To : %s\n %s", mail_to, email_tmp->data);
	}
	
}


struct state
{
	int state;
	int length;
};

const char *readlinefp_cb (void **buf, int *len, void *arg)
{

  int octets;
  email_object_t *object = (email_object_t *) arg;
  struct state *state;

  if (*buf == NULL)
		*buf = malloc(sizeof(struct state));

  state = *buf;
  
  if (len == NULL)
    {
      //rewind ((FILE *) arg);
	  state->state = 0;
	  state->length = object->data_size;
      return NULL;
    }

  switch (state->state) {
		case 0:
			state->state = 1;
			*len = state->length;
			break;
		default:
			*len = 0;
			break;
  }
  return object->data;
}

/*
 * send_message()
 *
 * Uploads the message to the server
 */
static int send_message(email_object_t *object)
{
	smtp_session_t session;
  	smtp_message_t message;
  	smtp_recipient_t recipient;
	char *host = NULL;
  	char *from = NULL;
  	char *subject = NULL;
	struct sigaction sa;
	const smtp_status_t *status;

	session = smtp_create_session ();
	message = smtp_add_message (session);
	
	/* NB.  libESMTP sets timeouts as it progresses through the protocol.
     In addition the remote server might close its socket on a timeout.
     Consequently libESMTP may sometimes try to write to a socket with
     no reader.  Ignore SIGPIPE, then the program doesn't get killed
     if/when this happens. */
	sa.sa_handler = SIG_IGN;
	sigemptyset (&sa.sa_mask);
	sa.sa_flags = 0;
	sigaction (SIGPIPE, &sa, NULL);

  /* Set the host running the SMTP server.  LibESMTP has a default port
     number of 587, however this is not widely deployed so the port
     is specified as 25 along with the default MTA host. */
  	smtp_set_server (session, "smtp:25");
 
	//smtp_set_message_fp(message, fp);
	smtp_set_messagecb(message, readlinefp_cb, (void *)object);	
	smtp_add_recipient (message, "stelios@cs.columbia.edu");
	/* Initiate a connection to the SMTP server and transfer the
     message. */
  if (!smtp_start_session (session))
    {
      char buf[128];

      fprintf (stderr, "SMTP server problem %s\n",
               smtp_strerror (smtp_errno (), buf, sizeof buf));
	  
	    
	  smtp_destroy_session (session);
	  return -1;
    }
  else
    {
      /* Report on the success or otherwise of the mail transfer.
       */
      status = smtp_message_transfer_status (message);
      printf ("%d %s", status->code,
              (status->text != NULL) ? status->text : "\n");
      //smtp_enumerate_recipients (message, print_recipient_status, NULL);
    }

    smtp_destroy_session (session);
	return 0;
}

/*
 * email_upload_objects()
 *
 * This function is responsible for uploading objects to a MTA
 *
 * There are a number of mechanisms that can be used to upload. 
 *
 * FIFO, LIFO, and shortest object first (SJF).
 *
 * returns 0 on success, -1 otherwise
 */
int email_upload_objects()
{
	GList *iterator, *sorted;
	email_object_t *email_tmp;
	
	sorted = g_list_copy(email_objects_buffer);

	if (!sorted) {
		fprintf(stderr, "email_upload_objects: No memory for sorted copy\n");
		return -1;
	}

	/*
	 * Choose mechanism to upload email objects
	 */
	switch(upload_mechanism) {
		/* Shortest email first */
		case SJF:
			sorted = g_list_sort(sorted,
							(GCompareFunc)email_compare_size_func);
			break;		
		default: 
			sorted = g_list_sort(sorted,
							(GCompareFunc)email_compare_size_func);
 
	}

	for (iterator = sorted; iterator != NULL; 
								iterator = iterator->next) {
		email_tmp = (email_object_t *) iterator->data;

		/*
		 * Upload email
		 */
		printf("Email %s is of size %d\n", email_tmp->mail_to,
						email_tmp->data_size);

		if ((send_message(email_tmp)) == 0) {
			/*message sent, delete from list*/
			//email_delete_object(email_tmp->md5);

			/*TODO:Send pager message*/
		} else {
			fprintf(stderr, "email_upload_objects: Couldn't send message!\n");
		}
	} 
	
	return 0; 
}

