/******************************************************************************
 * Multicast.c
 *
 * The multicast daemon for sevends
 *
 * Author:
 * Stelios Sidiroglou-Douskos <stelios@cs.columbia.edu>
 * Department of Computer Science - Columbia University
 *
 * Copyright (c) 2003 by Columbia University, all rights reserved.
 *****************************************************************************/

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>             /* close */
#include <string.h>             /*memcpy */
#include <pthread.h>            /* Threading lib */
#include <glib.h>

#include "main.h"
#include "multicast.h"
#include "xmlparser.h"
#include "glimpse_search_engine.h"
#include "cached_objects_hash.h"
#include "parser.h"
#include "swish_search_engine.h"
#include "query_report_buffer.h"

#define MULTICAST_GROUP "225.0.155.155"
#define ERROR "Error:multicast_server:"
#define MAX_MSG 10000
#define MSG_LEN	10000
//#define LOCAL_DEBUG 1 /*If defined, it will not see requests from local address */

/******************************* DECLARATIONS *********************************/

static void process_request(void *ptr);
extern int mcast_port_cfg;
extern int unicast_port_cfg;
extern char my_ip_address[16];

/******************************* DECLARATIONS *********************************/


/*
 * multicast_keyword()
 * Sends the keyword to the multicast network.
 * returns -1 on error, 0 on success
 */
int multicast_keyword(void *ptr)
{
    unsigned char ttl = 10;      /*FIXME: KEEP AT 1 FOR NOW */
    struct sockaddr_in cliAddr, servAddr;
    struct hostent *h;
    int rc;                     /* receive count */
    int sd;                     /* socket descriptor */
	char xml[MSG_LEN];
	int	send_count;
	char *keyword;

	keyword = (char *) ptr;
    h = gethostbyname(MULTICAST_GROUP);

	if (h == NULL) {
        fprintf(stderr, "Error:multicast_keyword unknown host '%s'\n",
               MULTICAST_GROUP);
        return -1;
    }

    servAddr.sin_family = h->h_addrtype;
    memcpy((char *) &servAddr.sin_addr.s_addr, h->h_addr_list[0],
           h->h_length);
    servAddr.sin_port = htons(mcast_port_cfg);

    /* check if dest address is multicast */
    if (!IN_MULTICAST(ntohl(servAddr.sin_addr.s_addr))) {
        fprintf
            (stderr, 
			 	"Error:multicast_keyword: given address '%s' is not multicast \n",
             inet_ntoa(servAddr.sin_addr));
        return -1;
    }

    /* create socket */
    sd = socket(AF_INET, SOCK_DGRAM, 0);
    if (sd < 0) {
        fprintf(stderr, "Error:multicast_keyword : cannot open socket\n");
        return -1;
    }

    /* bind any port number */
    cliAddr.sin_family = AF_INET;
    cliAddr.sin_addr.s_addr = htonl(INADDR_ANY);
    cliAddr.sin_port = htons(0);

    if (bind(sd, (struct sockaddr *) &cliAddr, sizeof(cliAddr)) < 0) {
        perror("bind");
        return -1;
    }

    if (setsockopt(sd, IPPROTO_IP, IP_MULTICAST_TTL, &ttl, sizeof(ttl)) <
        0) {
        printf("Error:multicast_keyword  : cannot set ttl = %d \n", ttl);
        return -1;
    }

	bzero(xml, sizeof(xml));    	
	/* 
	 * Create xml from keyword. xml is allocated mem from 
	 * create_query_xml so free xml after we're done 
	 */
	if (create_query_xml(keyword, &(xml[0])) < 0) {
		fprintf(stderr, "multicast_keyword:: error creating query\n");
		return -1;
	}
	
	for (send_count = 0; send_count < get_qrb_num_of_broadcasts(); send_count++) {
	
		printf("sending data on multicast group '%s' (%s)\n : %s \n",
           		h->h_name, inet_ntoa(*(struct in_addr *) h->h_addr_list[0]),
		   		&(xml[0]));
	
		rc = sendto(sd, xml, strlen(xml), 0,
                (struct sockaddr *) &servAddr, sizeof(servAddr));

    	if (rc < 0) {
        	fprintf(stderr, "Error:multicast_keyword: sendto failed \n");
	        close(sd);
    	}

		sleep(get_qrb_broadcast_freq());
	}
	
    close(sd);
	
    return 0;
}


/*
 * multicast_server_init()
 * Method for initializing the multicast server
 */
int multicast_server_init(struct data_structures *data)
{
    int sd, rc, n, cliLen;
    struct ip_mreq mreq;
    struct sockaddr_in cliAddr, servAddr;
    struct in_addr mcastAddr;
    struct hostent *h;
    char msg[MAX_MSG];
	pthread_t request_thread;
	struct multicast_data mdata;
	struct in_addr local_ip_address; 
			
	printf("multicast_server_init()\n");

    /* get mcast address to listen to */
    h = gethostbyname(MULTICAST_GROUP);
    if (h == NULL) {
        printf("%s: unknown group '%s'\n", ERROR, MULTICAST_GROUP);
        return -1;
    }

#ifdef LOCAL_DEBUG	
	inet_aton(my_ip_address, &local_ip_address);
#endif
    memcpy(&mcastAddr, h->h_addr_list[0], h->h_length);

    /* check given address is multicast */
    if (!IN_MULTICAST(ntohl(mcastAddr.s_addr))) {
        printf(" : given address '%s' is not multicast\n",
               inet_ntoa(mcastAddr));
        return -1;
    }

    /* create socket */
    sd = socket(AF_INET, SOCK_DGRAM, 0);
    if (sd < 0) {
        printf("%s : cannot create socket\n", ERROR);
        return -1;
    }

    /* bind port */
    servAddr.sin_family = AF_INET;
    servAddr.sin_addr.s_addr = htonl(INADDR_ANY);
    servAddr.sin_port = htons(mcast_port_cfg);
    if (bind(sd, (struct sockaddr *) &servAddr, sizeof(servAddr)) < 0) {
        printf("%s : cannot bind port %d \n", ERROR, mcast_port_cfg);
        return -1;
    }

    /* join multicast group */
    mreq.imr_multiaddr.s_addr = mcastAddr.s_addr;
    mreq.imr_interface.s_addr = htonl(INADDR_ANY);

    rc = setsockopt(sd, IPPROTO_IP, IP_ADD_MEMBERSHIP,
                    (void *) &mreq, sizeof(mreq));

    if (rc < 0) {
        printf("%s : cannot join multicast group '%s'", ERROR,
               inet_ntoa(mcastAddr));
		perror("Multicast: IP_ADD_MEMBERSHIP");
        return -1;
    } else {
        printf("%s : listening to mgroup %s:%d\n",
               MULTICAST_GROUP, inet_ntoa(mcastAddr), mcast_port_cfg);


        /* infinite server loop */
        while (1) {
			bzero(&msg, sizeof(msg));
			
            cliLen = sizeof(cliAddr);
            n = recvfrom(sd, msg, MAX_MSG, 0,
                         (struct sockaddr *) &cliAddr, &cliLen);

            if (n < 0) {
                printf("%s : cannot receive data\n", ERROR);
                continue;
            }

            printf("Received Multicast query from %s:%d on %s : %s\n",
                   inet_ntoa(cliAddr.sin_addr), ntohs(cliAddr.sin_port),
                   MULTICAST_GROUP, msg);
#ifdef LOCAL_DEBUG	
			/*FIXME: If from our address reject */
			if (cliAddr.sin_addr.s_addr == local_ip_address.s_addr) {
				printf(">>>>>>>> Request from our address... continuing\n");
				continue;
			}
#endif
			/* copy data over */
			//mdata.data = data;
			//mdata.msg = &(msg[0]);        
			mdata.msg = malloc(MAX_MSG);
			memcpy(mdata.msg, &(msg[0]), MAX_MSG);
			mdata.cliAddr = cliAddr;
			
			/* create thread */
			if ((pthread_create(&request_thread,
                            NULL,
                            (void *(*)(void *)) process_request,
                            (void *) &mdata)) != 0) {
            	printf("Error -> ProxyServer: Error creating thread\n");
        	}

		}                       /* end of infinite server loop */
		
		return 0; /*ama erthei pote edw */
    }
}

/*
 * send_report_url()
 * 
 * return 0 on success, -1 otherwise
 */
static int send_report_url(char *path, struct sockaddr_in cliAddr)
{
	int sd;
	char xml_response[MSG_LEN];
	
	/* create socket */
    sd = socket(AF_INET, SOCK_DGRAM, 0);
    if (sd < 0) {
        printf("%s : cannot create socket\n", ERROR);
        return -1;
    }

	bzero(xml_response, sizeof(xml_response));
	/*
	 * create xml report based on url in cache
	 */
	if (create_response_xml(path, &(xml_response[0])) < 0) {
		printf("%s : cannot create xml response\n", ERROR);
	   	return -1;	
	}

	/*
	 * FIXME: Check this
	 */
	cliAddr.sin_port = htons(unicast_port_cfg);
	sendto(sd, &(xml_response[0]), strlen(xml_response), 0, 
					(struct sockaddr *) &cliAddr, sizeof(cliAddr));


	close(sd);

	return 0;
}

/*
 * send_report_keyword()
 */
static int send_report_keyword(char *xml, struct sockaddr_in cliAddr)
{
	int sd;
	
	/* create socket */
    sd = socket(AF_INET, SOCK_DGRAM, 0);
    if (sd < 0) {
        printf("%s : cannot create socket\n", ERROR);
        return -1;
    }

	printf("Sending to %s\n", inet_ntoa(cliAddr.sin_addr));
	/*
	 * FIXME: Check this
	 */
	cliAddr.sin_port = htons(unicast_port_cfg);
	sendto(sd,xml, strlen(xml), 0, 
					(struct sockaddr *) &cliAddr, sizeof(cliAddr));

	close(sd);

	return 0;
}

/*
 * service_url_request()
 * 
 * Service requests that look like they're requests for urls
 * 
 */
static void service_url_request(char *request, struct sockaddr_in cliAddr) 
{
	gpointer value = NULL;
	char *path;
	/*If we find the requested url, send report*/
	value = is_url_in_hash(request);

	if (value == NULL) {
		printf("service_url_request::HASH LOOKUP NOT FOUND %s\n", (char *)value);
        return;
	} else {
        printf("service_url_request::HASH LOOKUP IS %s\n", (char *)value);
	}

	path = (char *) value;

	send_report_url(path, cliAddr);
	
}

/*
 * service_keyword_request()
 *
 * Service requests for keywords... Send glimpse results to querier
 */
static void service_keyword_request(char *request, struct sockaddr_in cliAddr)
{
	/*
	 * Query glimpse with keyword, create xml report with results and 
	 * send them back to querier.
	 */
	char xml[MSG_LEN];
		
	bzero(xml, sizeof(xml));
	
	if (swish_get_report(request, &(xml[0])) < 0) {
		fprintf(stderr, "service_keyword_request:Error getting report %s\n",
						request);
		return;
	}

	
	send_report_keyword(&(xml[0]), cliAddr);
	bzero(xml, sizeof(xml));	

	return;	
}

/*
 * process_request()
 * this method processes reports that come in from the multicast 
 * network. These are the general queries sent by other 7ds nodes.
 */
static void process_request(void *ptr)
{
    /*process queries */
	struct multicast_data *mdata;
	char request[MSG_LEN];
	
	mdata = (struct multicast_data *) ptr;

	bzero(request, sizeof(request));
	/*
	 * parse the xml request. request should now either be a string that 
	 * contains the request or a URL in the form of http://www.xxx.xxx
	 */
	if (parse_multicast_xml(mdata->msg, &(request[0])) < 0) {
		fprintf(stderr, "process_request:: Error parsing multicast xml\n");
		//goto clean_request;
		return;
	}

	if (is_request_url(request)) {
			service_url_request(request, mdata->cliAddr);
	} else {
			service_keyword_request(request, mdata->cliAddr);
	}

	free(mdata->msg); /*malloc in multicast server init*/

	return;
	
clean_request:
	/*
	 * request is allocated by parse_multicast_xml so free it here
	 */
	if (request != NULL)
		free(request);	

}
