/******************************************************************************
 * referer_hash.c
 * 
 * Hashtable of referers... 
 * Author:
 * Stelios Sidiroglou-Douskos <stelios@cs.columbia.edu>
 * Department of Computer Science - Columbia University
 *
 * Copyright (c) 2003 by Columbia University, all rights reserved.
 * ***************************************************************************/

/*Glib libs*/
#include <pthread.h>            /* Threading lib */
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <glib.h>
#include <string.h>
#include "referer_hash.h"

GHashTable *referer_hash;
static long referer_create_timer();
static void referer_put_in_hash(char *referer, long *time);
long max_referer_time = 120; /* In seconds */

/*
 * referer_hash_init()
 * Initialize hash table
 */
void referer_hash_init(void)
{
	 referer_hash = NULL;
	 referer_hash = g_hash_table_new(g_str_hash, g_str_equal);
}

/*
 * referer_hash_destroy()
 * Destroy Hash Table, free memory
 */
void referer_hash_destroy(void)
{
	g_hash_table_destroy(referer_hash);
}

/*
 * referer_is_key_in_hash()
 * If referer is in the hash, check to see if the association is expired.
 * If it's expired, remove from the table. 
 * 
 * Returns 1 if the association is valid, 0 otherwise.
 */
int referer_is_key_in_hash(char *referer)
{
	gpointer value = NULL;
    pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
	long referer_time;
	long current_time;	
	pthread_mutex_lock(&mutex);
	
	value = g_hash_table_lookup(referer_hash,		/*table */
                                referer);			/*key */
    pthread_mutex_unlock(&mutex);

	if (value == NULL) {
		return 0;
	} else {
		/* If it's there, check to see if it's timedout */
		current_time = referer_create_timer();
		referer_time = (long) value;
		printf("Referer_hash::is_key_in_hash"
					" current_time:%ld  referer_time:%ld\n", 
						current_time, referer_time);

		if ((current_time - referer_time) > max_referer_time) {
			referer_remove_from_hash(referer);
			return 0;
		}
	}

	return 1;
}

/*
 * referer_create_referer()
 *
 * Calls create timer and inserts pair into hashtable
 */
void referer_create_referer(char *referer) 
{
	long timer;

	timer = referer_create_timer();
	referer_put_in_hash(referer, &timer);
}

/*
 * cob_put_in_cache()
 * Places the path of the file under the url key
 */
static void referer_put_in_hash(char *referer, long *time)
{
	pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;

    printf("Referer_hash::Inserting -> referer %s\n time %ld\n", 
													referer, time);

    pthread_mutex_lock(&mutex);
    g_hash_table_insert(referer_hash, referer, time);
    pthread_mutex_unlock(&mutex);
}

/*
 * referer_create_timer()
 * Creates a timer value in seconds and returns a long with
 * the number.
 */
static long referer_create_timer()
{
	GTimeVal result;

	g_get_current_time(&result);
	
	return result.tv_sec;	
}

/*
 * referer_get_number_in_hash()
 * Returns the number of elements in the hash
 */
unsigned int referer_get_number_in_hash(void)
{
	pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
	unsigned int size;
	 
	pthread_mutex_lock(&mutex);
    size = g_hash_table_size(referer_hash);
	pthread_mutex_unlock(&mutex);
	
	return size;
}

/*
 * referer_remove_from_hash()
 * Remove the value with the key passed from hash
 */
void referer_remove_from_hash(char *referer)
{
	pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
	pthread_mutex_lock(&mutex);
	g_hash_table_remove(referer_hash, referer);	
	pthread_mutex_unlock(&mutex);
}


