/******************************************************************************
 *
 * swish_search_engine.c
 *
 * A search engine interface using the swish library. 
 *
 * http://swish-e.org/
 * 
 * Author:
 * Stelios Sidiroglou-Douskos <stelios@cs.columbia.edu>
 * Department of Computer Science - Columbia University
 *
 * Copyright (c) 2003 by Columbia University, all rights reserved.
 * ***************************************************************************/

#include <pthread.h>            /* Threading lib */
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>

#ifdef ARM
	#include "libs/swish-e.h"
#else 
	#include <swish-e.h>
#endif

#include "xmlparser.h"
#include <string.h>

#define INDEX "docs.index"

/*
 * create_swish_xml_report()
 * 
 * Returns 0 on success, -1 on failure
 */
static int create_swish_xml_report(xmlNodePtr root, xmlNodePtr query_cur, 
								SW_RESULT result)
{
	time_t time;
	char buf[64];
	
	query_cur = xmlNewChild(root, NULL, 
					(const xmlChar *) "Report", NULL);

	/*
	 * Add local path
	 */
	create_xml_local_url(query_cur, query_cur, 
						SwishResultPropertyStr(result, "swishdocpath"));

	/*
	 * Add title
	 */
	create_xml_title(query_cur, query_cur, 
						SwishResultPropertyStr(result, "swishtitle"));

	
	/*
   	 * Add rank
	 */
	create_xml_rank(query_cur, query_cur,
							SwishResultPropertyStr(result, "swishrank"));

	/*
	 * Add size
	 */
	create_xml_size(query_cur, query_cur,
							SwishResultPropertyStr(result, "swishdocsize"));


	/*
	 * Last modification Date
	 */
	time = SwishResultPropertyULong(result, "swishlastmodified");
	sprintf(buf, "%lu", time);
	create_xml_date(query_cur, query_cur, buf);
	
	return 0;
}


/*
 * swish_get_report()
 *
 * Returns 0 on success, -1 on failure
 */
int swish_get_report(char *keyword, char *xml)
{
	SW_SEARCH   search = NULL; /* holds search parameters */
    SW_RESULTS  results = NULL; /* results handle -- holds list of results */
	SW_HANDLE   swish_handle = NULL; /* Database handle */
	int hits;
	SW_RESULT result;

	/*xml stuff*/
	xmlDocPtr doc;
	xmlNodePtr cur;
	xmlNodePtr query_cur;
	xmlNodePtr root;
	xmlChar *mem;
	int size;

	swish_handle = SwishInit(INDEX);

	/* Check for errors after every call */
    
    if ( SwishError( swish_handle ) ) {
		goto swish_error;  
	}

	/* 
	 * Here's a short-cut to searching that creates a search object and 
	 * searches at the same time 
	 */

    results = SwishQuery( swish_handle, keyword);
    
    if ( SwishError( swish_handle ) ) {
		goto swish_error;        
	} else {
   		/*
	 	 * Now check the results...
	 	 */
        printf( "Testing SW_ResultsToSW_HANDLE() = '%s'\n",
            SW_ResultsToSW_HANDLE( results ) == swish_handle ? "OK" : "Not OK" );
        
		hits = SwishHits(results);

		if (hits == 0) {
			printf("No results for query %s\n", keyword);
			Free_Results_Object(results);
			Free_Search_Object(search);
			SwishClose(swish_handle);
			return  -1;
		}

		printf("# Total Results: %d\n", hits);
	
		/*
		 * Now set up xml
		 */
		
		/*
	 	 * Create XML doc
	 	 */
		doc = xmlNewDoc((const xmlChar *) "1.0");
	
		if (doc == NULL ) {
			fprintf(stderr,
					"get_report:: Document not created successfully.\n");
			goto swish_error;
		}

		root = xmlNewNode(NULL, (const xmlChar *) XML_ROOT);
	
		xmlDocSetRootElement(doc, root);
	
		if (root == NULL) {
			fprintf(stderr,
						"get_report:: Root node not created\n");
			goto swish_error;

		}
			
		/*
	 	 * Add query structure
	 	 */
		if ((cur = create_xml_query(root, query_cur, keyword)) == NULL) {
			fprintf(stderr, "get_report:: Couldn't creat xml doc\n");
			goto swish_error;
		}

		/*
		 * Add query name to xml report
		 */
		if (create_xml_query_name(cur, query_cur, keyword) < 0) {
			fprintf(stderr, "get_report:: Couldn't creat xml doc\n");
			goto swish_error;
		}
	
		/*
		 * Step through results
		 */
		while ( (result = SwishNextResult( results )) ) {
			
	        printf("Path: %s\n  Rank: %lu\n  Size: %lu\n  Title: %s\n  	Index: %s\n  Modified: %s\n  Record #: %lu\n  File   #: %lu\n  Date: %lu\n Description:	%s\n",
            SwishResultPropertyStr   ( result, "swishdocpath" ),
            SwishResultPropertyULong ( result, "swishrank" ),
            SwishResultPropertyULong ( result, "swishdocsize" ),
            SwishResultPropertyStr   ( result, "swishtitle"),
            SwishResultPropertyStr   ( result, "swishdbfile" ),
            SwishResultPropertyStr   ( result, "swishlastmodified" ),
            SwishResultPropertyULong ( result, "swishreccount" ),  
			SwishResultPropertyULong ( result, "swishfilenum" ),
			SwishResultPropertyULong   ( result, "swishlastmodified" ),
			SwishResultPropertyStr	 ( result, "swishdescription"));

			/*
			 * Create Report
			 */	
			if (create_swish_xml_report(cur, query_cur, result) < 0) {
				fprintf(stderr, "get_report:: Couldn't creat xml doc\n");
				goto swish_error;
			}
			
		}

		
        Free_Results_Object( results );
    }

	/* Copy to xml */
	xmlDocDumpMemory(doc, &mem, &size);	
	strncpy(xml, mem, size); 


	/*
	 * Cleanup
	 */
	Free_Search_Object(search);
	SwishClose(swish_handle);
	xmlFreeDoc(doc);

	return 0;

swish_error:
	fprintf(stderr, "swish_get_report:: Error connecting to swish. %s\n",
							SwishErrorString(swish_handle));
	SwishClose(swish_handle); /* free mem used */
	xmlFreeDoc(doc);
	return -1;

}
